""" Trailblazer API for cg """ ""
import datetime as dt
import logging
import shutil
from pathlib import Path
from typing import List

import click
import ruamel.yaml
from trailblazer.mip.start import MipCli
from trailblazer.store import api, models, Store
from trailblazer.mip import files, trending

LOG = logging.getLogger(__name__)


class TrailblazerAPI(Store):
    """Interface to Trailblazer for `cg`."""

    parse_sampleinfo = staticmethod(files.parse_sampleinfo)

    def __init__(self, config: dict):
        super(TrailblazerAPI, self).__init__(
            config["trailblazer"]["database"], families_dir=config["trailblazer"]["root"]
        )
        self.mip_cli = MipCli(
            script=config["trailblazer"]["script"],
            pipeline=config["trailblazer"]["pipeline"],
            conda_env=config["trailblazer"]["conda_env"],
        )
        self.mip_config = config["trailblazer"]["mip_config"]

    def mark_analyses_deleted(self, case_id: str):
        """ mark analyses connected to a case as deleted """
        for old_analysis in self.analyses(family=case_id):
            old_analysis.is_deleted = True
        self.commit()

    def add_pending_analysis(self, case_id: str, email: str = None) -> models.Analysis:
        """ Add analysis as pending"""
        self.add_pending(case_id, email)

    @staticmethod
    def get_sampleinfo(analysis: models.Analysis) -> str:
        """Get the sample info path for an analysis."""
        raw_data = ruamel.yaml.safe_load(Path(analysis.config_path).open())
        data = files.parse_config(raw_data)
        return data["sampleinfo_path"]

    @staticmethod
    def parse_qcmetrics(data: dict) -> dict:
        """Call internal Trailblazer MIP API."""
        return files.parse_qcmetrics(data)

    def is_analysis_ongoing(self, case_id: str) -> bool:
        """Call internal Trailblazer API"""
        return self.is_latest_analysis_ongoing(case_id=case_id)

    def is_analysis_failed(self, case_id: str) -> bool:
        """Call internal Trailblazer API"""
        return self.is_latest_analysis_failed(case_id=case_id)

    def is_analysis_completed(self, case_id: str) -> bool:
        """Call internal Trailblazer API"""
        return self.is_latest_analysis_completed(case_id=case_id)

    def get_analysis_status(self, case_id: str) -> str:
        """Call internal Trailblazer API"""
        return self.get_latest_analysis_status(case_id=case_id)

    def has_analysis_started(self, case_id: str) -> bool:
        """Check if analysis has started"""
        statuses = ("ongoing", "failed", "completed")
        get_analysis_status = {
            "ongoing": self.is_analysis_ongoing,
            "failed": self.is_analysis_failed,
            "completed": self.is_analysis_completed,
        }
        for status in statuses:
            has_started = get_analysis_status[status](case_id=case_id)
            if has_started:
                return has_started
        return False

    def delete_analysis(
        self, family: str, date: dt.datetime, yes: bool = False, dry_run: bool = False
    ):
        """Delete the analysis output."""
        if self.analyses(family=family, temp=True).count() > 0:
            raise ValueError("analysis for family already running")
        analysis_obj = self.find_analysis(family, date, "completed")
        assert analysis_obj.is_deleted is False
        analysis_path = Path(analysis_obj.out_dir).parent

        if yes or click.confirm(f"Do you want to remove {analysis_path}?"):

            if not dry_run:
                shutil.rmtree(analysis_path, ignore_errors=True)
                analysis_obj.is_deleted = True
                self.commit()

    @staticmethod
    def get_trending(mip_config_raw: str, qcmetrics_raw: str, sampleinfo_raw: dict) -> dict:
        """Get trending data for a MIP analysis"""
        return trending.parse_mip_analysis(
            mip_config_raw=mip_config_raw,
            qcmetrics_raw=qcmetrics_raw,
            sampleinfo_raw=sampleinfo_raw,
        )

    def get_family_root_dir(self, family_id: str):
        """Get path for a case"""
        return Path(self.families_dir) / family_id

    def get_latest_logged_analysis(self, case_id: str):
        """Get the the analysis with the latest logged_at date"""
        return self.analyses(family=case_id).order_by(models.Analysis.logged_at.desc())

    @staticmethod
    def get_sampleinfo_date(data: dict) -> str:
        """Get date from a sampleinfo """
        return files.get_sampleinfo_date(data)



def test_microbial_sample_to_dict(microbial_store):

    # GIVEN a store with a Microbial sample
    assert microbial_store.microbial_samples().count() > 0

    # WHEN running to dict on that sample
    a_dict = microbial_store.microbial_samples().first().to_dict(order=True)

    # THEN you should get a dictionary with
    assert a_dict["id"]
    assert a_dict["internal_id"]
    assert a_dict["name"]
    assert a_dict["application_version_id"]
    assert a_dict["microbial_order_id"]
    assert a_dict["created_at"]
    assert a_dict["organism"]
    assert a_dict["organism_id"]
    assert a_dict["reference_genome"]
    assert a_dict["priority"]
    assert a_dict["reads"]
    assert a_dict["comment"]
    assert a_dict["microbial_order"]
    assert a_dict["application"]
    assert a_dict["application_version"]


def test_order_sample_to_dict(microbial_store):

    # GIVEN a store with a Microbial sample
    assert microbial_store.microbial_orders().count() > 0

    # WHEN running to dict on that sample
    a_dict = microbial_store.microbial_orders().first().to_dict(samples=True)

    # THEN you should get a dictionary with
    assert a_dict["id"]
    assert a_dict["internal_id"]
    assert a_dict["name"]
    assert a_dict["ticket_number"]
    assert a_dict["comment"]
    assert a_dict["created_at"]
    assert a_dict["updated_at"]
    assert a_dict["ordered_at"]
    assert a_dict["customer_id"]
    assert a_dict["microbial_samples"]
    assert a_dict["customer"]



"""
    Tests for GenotypeAPI
"""

import subprocess
import pytest

from cg.apps.gt import GenotypeAPI
from cg.exc import CaseNotFoundError


def test_instatiate(genotype_config):

    """Test to instantiate a genotype api"""

    # GIVEN a genotype api with a config

    # WHEN instantiating a genotype api
    genotype_api = GenotypeAPI(genotype_config)

    # THEN assert that the adapter was properly instantiated
    assert genotype_api.genotype_config == genotype_config["genotype"]["config_path"]
    assert genotype_api.genotype_binary == genotype_config["genotype"]["binary_path"]


def test_export_sample(genotypeapi, mocker):

    """Test that get_trending calls the genotype API with correct command."""

    # GIVEN a genotypeapi api and argument days
    days = 20

    # WHEN running get_trending
    mocker.patch.object(subprocess, "check_output")
    genotypeapi.export_sample(days=days)

    # THEN assert subprocess is running the GenotypeAPI with correct command
    call = ["gtdb", "--config", "config/path", "export-sample", "-d", str(days)]
    subprocess.check_output.assert_called_with(call)


def test_export_sample_no_output(genotypeapi, mocker):

    """Test to get genotype via the api"""

    # GIVEN a genotypeapi api

    # WHEN export_sample is calling the GenotypeAPI witch is returning a empty string
    mocker.patch.object(subprocess, "check_output")
    subprocess.check_output.return_value = b""

    # THEN assert CaseNotFoundError
    with pytest.raises(CaseNotFoundError):
        genotypeapi.export_sample(days="")



import gzip
import logging
import re
from pathlib import Path
from typing import List, Any
from ruamel.yaml import safe_load

from requests.exceptions import HTTPError

from cg.apps import tb, hk, scoutapi, lims
from cg.apps.pipelines.fastqhandler import BaseFastqHandler
from cg.store import models, Store


class AnalysisAPI:
    """The pipelines are accessed through Trailblazer but cg provides additional conventions and
    hooks into the status database that makes managing analyses simpler"""

    def __init__(
        self,
        db: Store,
        hk_api: hk.HousekeeperAPI,
        lims_api: lims.LimsAPI,
        yaml_loader=safe_load,
        path_api=Path,
        logger=logging.getLogger(__name__),
    ):
        self.db = db
        self.hk = hk_api
        self.lims = lims_api
        self.yaml_loader = yaml_loader
        self.pather = path_api
        self.LOG = logger

    def check(self, family_obj: models.Family):
        """Check stuff before starting the analysis."""
        flowcells = self.db.flowcells(family=family_obj)
        statuses = []
        for flowcell_obj in flowcells:
            self.LOG.debug(f"{flowcell_obj.name}: checking flowcell")
            statuses.append(flowcell_obj.status)
            if flowcell_obj.status == "removed":
                self.LOG.info(f"{flowcell_obj.name}: requesting removed flowcell")
                flowcell_obj.status = "requested"
            elif flowcell_obj.status != "ondisk":
                self.LOG.warning(f"{flowcell_obj.name}: {flowcell_obj.status}")
        return all(status == "ondisk" for status in statuses)

    @staticmethod
    def fastq_header(line):
        """handle illumina's two different header formats
        @see https://en.wikipedia.org/wiki/FASTQ_format

        @HWUSI-EAS100R:6:73:941:1973#0/1

            HWUSI-EAS100R   the unique instrument name
            6   flowcell lane
            73  tile number within the flowcell lane
            941     'x'-coordinate of the cluster within the tile
            1973    'y'-coordinate of the cluster within the tile
            #0  index number for a multiplexed sample (0 for no indexing)
            /1  the member of a pair, /1 or /2 (paired-end or mate-pair reads only)

        Versions of the Illumina pipeline since 1.4 appear to use #NNNNNN
        instead of #0 for the multiplex ID, where NNNNNN is the sequence of the
        multiplex tag.

        With Casava 1.8 the format of the '@' line has changed:

        @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG

            EAS139  the unique instrument name
            136     the run id
            FC706VJ     the flowcell id
            2   flowcell lane
            2104    tile number within the flowcell lane
            15343   'x'-coordinate of the cluster within the tile
            197393  'y'-coordinate of the cluster within the tile
            1   the member of a pair, 1 or 2 (paired-end or mate-pair reads only)
            Y   Y if the read is filtered, N otherwise
            18  0 when none of the control bits are on, otherwise it is an even number
            ATCACG  index sequence


        TODO: add unit test
        """

        rs = {"lane": None, "flowcell": None, "readnumber": None}

        parts = line.split(":")
        if len(parts) == 5:  # @HWUSI-EAS100R:6:73:941:1973#0/1
            rs["lane"] = parts[1]
            rs["flowcell"] = "XXXXXX"
            rs["readnumber"] = parts[-1].split("/")[-1]
        if len(parts) == 10:  # @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG
            rs["lane"] = parts[3]
            rs["flowcell"] = parts[2]
            rs["readnumber"] = parts[6].split(" ")[-1]
        if len(parts) == 7:  # @ST-E00201:173:HCLCGALXX:1:2106:22516:34834/1
            rs["lane"] = parts[3]
            rs["flowcell"] = parts[2]
            rs["readnumber"] = parts[-1].split("/")[-1]

        return rs

    def link_sample(self, fastq_handler: BaseFastqHandler, sample: str, case: str):
        """Link FASTQ files for a sample."""
        file_objs = self.hk.files(bundle=sample, tags=["fastq"])
        files = []

        for file_obj in file_objs:
            # figure out flowcell name from header
            with gzip.open(file_obj.full_path) as handle:
                header_line = handle.readline().decode()
                header_info = self.fastq_header(header_line)

            data = {
                "path": file_obj.full_path,
                "lane": int(header_info["lane"]),
                "flowcell": header_info["flowcell"],
                "read": int(header_info["readnumber"]),
                "undetermined": ("_Undetermined_" in file_obj.path),
            }
            # look for tile identifier (HiSeq X runs)
            matches = re.findall(r"-l[1-9]t([1-9]{2})_", file_obj.path)
            if len(matches) > 0:
                data["flowcell"] = f"{data['flowcell']}-{matches[0]}"
            files.append(data)

        fastq_handler.link(case=case, sample=sample, files=files)



""" This file groups all tests related to microsalt start creation """

from cg.cli.workflow.microsalt.base import run

EXIT_SUCCESS = 0


def test_no_arguments(cli_runner, base_context):
    """Test command without any options"""

    # GIVEN

    # WHEN dry running without anything specified
    result = cli_runner.invoke(run, obj=base_context)

    # THEN command should mention missing arguments
    assert result.exit_code != EXIT_SUCCESS


def test_dry_arguments(cli_runner, base_context, microbial_order_id, queries_path):
    """Test command dry """

    # GIVEN

    # WHEN dry running without anything specified
    result = cli_runner.invoke(run, ["--dry", microbial_order_id], obj=base_context)

    # THEN command should mention missing arguments
    assert result.exit_code == EXIT_SUCCESS

    outfilename = queries_path / microbial_order_id
    outfilename = outfilename.with_suffix(".json")
    assert f"/bin/true --parameters {outfilename}" in result.output



import logging

from subprocess import CalledProcessError
import subprocess

from alchy import Manager
from genotype.store import api, models
from genotype.load.vcf import load_vcf
from cg.exc import CaseNotFoundError

LOG = logging.getLogger(__name__)


class GenotypeAPI(Manager):

    """Interface with Genotype app.

    The config should contain a 'genotype' key:

        { 'database': 'mysql://localhost:3306/database' }
    """

    def __init__(self, config: dict):
        alchy_config = dict(SQLALCHEMY_DATABASE_URI=config["genotype"]["database"])
        super(GenotypeAPI, self).__init__(config=alchy_config, Model=models.Model)

        self.genotype_config = config["genotype"]["config_path"]
        self.genotype_binary = config["genotype"]["binary_path"]
        self.base_call = [self.genotype_binary, "--config", self.genotype_config]

    def upload(self, bcf_path: str, samples_sex: dict, force: bool = False):
        """Upload genotypes for a family of samples."""
        snps = api.snps()
        analyses = load_vcf(bcf_path, snps)
        for analysis_obj in analyses:
            LOG.debug("loading VCF genotypes for sample: %s", analysis_obj.sample_id)
            is_saved = api.add_analysis(self, analysis_obj, replace=force)
            if is_saved:
                LOG.info("loaded VCF genotypes for sample: %s", analysis_obj.sample_id)
            else:
                LOG.warning("skipped, found previous analysis: %s", analysis_obj.sample_id)

            if is_saved or force:
                analysis_obj.sex = samples_sex[analysis_obj.sample_id]["analysis"]
                analysis_obj.sample.sex = samples_sex[analysis_obj.sample_id]["pedigree"]
                self.commit()

    def export_sample(self, days: int = 0) -> str:
        """Export sample info."""
        trending_call = self.base_call[:]
        trending_call.extend(["export-sample", "-d", str(days)])
        try:
            LOG.info("Running Genotype API to get data.")
            LOG.debug(trending_call)
            output = subprocess.check_output(trending_call)
        except CalledProcessError as error:
            LOG.critical("Could not run command: %s", " ".join(trending_call))
            raise error
        output = output.decode("utf-8")
        # If sample not in genotype db, stdout of genotype command will be empty.
        if not output:
            raise CaseNotFoundError("samples not found in genotype db")
        return output

    def export_sample_analysis(self, days: int = 0) -> str:
        """Export analysis."""
        trending_call = self.base_call[:]
        trending_call.extend(["export-sample-analysis", "-d", str(days)])
        try:
            LOG.info("Running Genotype API to get data.")
            LOG.debug(trending_call)
            output = subprocess.check_output(trending_call)
        except CalledProcessError as error:
            LOG.critical("Could not run command: %s", " ".join(trending_call))
            raise error
        output = output.decode("utf-8")
        # If sample not in genotype db, stdout of genotype command will be empty.
        if not output:
            raise CaseNotFoundError("samples not found in genotype db")
        return output



"""Ordering module for intration"""
import datetime as dt
from typing import List

from cg.exc import OrderError
from cg.store import models


class StatusHandler:
    """Handles ordering data for the statusDB"""

    def __init__(self):
        self.status = None

    @staticmethod
    def group_cases(samples: List[dict]) -> dict:
        """Group samples in cases."""
        cases = {}
        for sample in samples:
            if sample["family_name"] not in cases:
                cases[sample["family_name"]] = []
            cases[sample["family_name"]].append(sample)
        return cases

    @staticmethod
    def pools_to_status(data: dict) -> dict:
        """Convert input to pools."""

        status_data = {"customer": data["customer"], "order": data["name"], "pools": []}

        # group pools
        pools = {}

        for sample in data["samples"]:
            name = sample["pool"]
            application = sample["application"]
            data_analysis = sample["data_analysis"]
            capture_kit = sample.get("capture_kit")

            if name not in pools:
                pools[name] = {}
                pools[name]["name"] = name
                pools[name]["applications"] = set()
                pools[name]["capture_kits"] = set()

            pools[name]["applications"].add(application)

            if capture_kit:
                pools[name]["capture_kits"].add(capture_kit)

        # each pool must only have one application type
        for pool in pools.values():

            applications = pool["applications"]
            pool_name = pool["name"]
            if len(applications) != 1:
                raise OrderError(f"different application in pool: {pool_name} - {applications}")

        # each pool must only have one capture kit
        for pool in pools.values():

            capture_kits = pool["capture_kits"]
            pool_name = pool["name"]

            if len(capture_kits) > 1:
                raise OrderError(f"different capture kits in pool: {pool_name} - {capture_kits}")

        for pool in pools.values():

            pool_name = pool["name"]
            applications = pool["applications"]
            application = applications.pop()
            capture_kits = pool["capture_kits"]
            capture_kit = None

            if len(capture_kits) == 1:
                capture_kit = capture_kits.pop()

            status_data["pools"].append(
                {
                    "name": pool_name,
                    "application": application,
                    "data_analysis": data_analysis,
                    "capture_kit": capture_kit,
                }
            )
        return status_data

    @staticmethod
    def samples_to_status(data: dict) -> dict:
        """Convert order input to status for fastq-only/metagenome orders."""
        status_data = {
            "customer": data["customer"],
            "order": data["name"],
            "samples": [
                {
                    "internal_id": sample.get("internal_id"),
                    "name": sample["name"],
                    "application": sample["application"],
                    "data_analysis": sample["data_analysis"],
                    "priority": sample["priority"],
                    "comment": sample.get("comment"),
                    "tumour": sample.get("tumour") or False,
                    "sex": sample.get("sex"),
                    "status": sample.get("status"),
                }
                for sample in data["samples"]
            ],
        }
        return status_data

    @staticmethod
    def microbial_samples_to_status(data: dict) -> dict:
        """Convert order input for microbial samples."""
        status_data = {
            "customer": data["customer"],
            "order": data["name"],
            "comment": data.get("comment"),
            "samples": [
                {
                    "priority": sample_data["priority"],
                    "name": sample_data["name"],
                    "internal_id": sample_data.get("internal_id"),
                    "organism_id": sample_data["organism"],
                    "comment": sample_data.get("comment"),
                    "reference_genome": sample_data["reference_genome"],
                    "application": sample_data["application"],
                    "data_analysis": sample_data["data_analysis"],
                }
                for sample_data in data["samples"]
            ],
        }
        return status_data

    @classmethod
    def cases_to_status(cls, data: dict) -> dict:
        """Convert order input to status interface input."""
        status_data = {
            "customer": data["customer"],
            "order": data["name"],
            "families": [],
        }
        cases = cls.group_cases(data["samples"])

        for case_name, case_samples in cases.items():
            values = set(sample.get("priority", "standard") for sample in case_samples)
            if len(values) > 1:
                raise ValueError(f"different 'priority' values: {case_name} - {values}")
            priority = values.pop()
            panels = set(panel for sample in case_samples for panel in sample.get("panels", set()))
            case = {
                "name": case_name,
                "priority": priority,
                "panels": list(panels),
                "samples": [
                    {
                        "application": sample["application"],
                        "capture_kit": sample.get("capture_kit"),
                        "comment": sample.get("comment"),
                        "data_analysis": sample.get("data_analysis"),
                        "father": sample.get("father"),
                        "from_sample": sample.get("from_sample"),
                        "internal_id": sample.get("internal_id"),
                        "mother": sample.get("mother"),
                        "name": sample["name"],
                        "sex": sample["sex"],
                        "status": sample.get("status"),
                        "time_point": sample.get("time_point"),
                        "tumour": sample.get("tumour") or False,
                    }
                    for sample in case_samples
                ],
            }

            status_data["families"].append(case)
        return status_data

    def store_cases(
        self,
        customer: str,
        order: str,
        ordered: dt.datetime,
        ticket: int,
        cases: List[dict],
    ) -> List[models.Family]:
        """Store cases and samples in the status database."""
        customer_obj = self.status.customer(customer)
        if customer_obj is None:
            raise OrderError(f"unknown customer: {customer}")
        new_families = []
        for case in cases:
            case_obj = self.status.find_family(customer_obj, case["name"])
            if case_obj:
                case_obj.panels = case["panels"]
            else:
                case_obj = self.status.add_family(
                    name=case["name"], panels=case["panels"], priority=case["priority"]
                )
                case_obj.customer = customer_obj
                new_families.append(case_obj)

            family_samples = {}
            for sample in case["samples"]:
                sample_obj = self.status.sample(sample["internal_id"])
                if sample_obj:
                    family_samples[sample["name"]] = sample_obj
                else:
                    new_sample = self.status.add_sample(
                        capture_kit=sample["capture_kit"],
                        comment=sample["comment"],
                        data_analysis=sample["data_analysis"],
                        from_sample=sample["from_sample"],
                        internal_id=sample["internal_id"],
                        name=sample["name"],
                        order=order,
                        ordered=ordered,
                        priority=case["priority"],
                        sex=sample["sex"],
                        ticket=ticket,
                        time_point=sample["time_point"],
                        tumour=sample["tumour"],
                    )
                    new_sample.customer = customer_obj
                    with self.status.session.no_autoflush:
                        application_tag = sample["application"]
                        new_sample.application_version = self.status.current_application_version(
                            application_tag
                        )
                    if new_sample.application_version is None:
                        raise OrderError(f"Invalid application: {sample['application']}")

                    family_samples[new_sample.name] = new_sample
                    self.status.add(new_sample)
                    new_delivery = self.status.add_delivery(destination="caesar", sample=new_sample)
                    self.status.add(new_delivery)

            for sample in case["samples"]:
                mother_obj = family_samples[sample["mother"]] if sample.get("mother") else None
                father_obj = family_samples[sample["father"]] if sample.get("father") else None
                with self.status.session.no_autoflush:
                    link_obj = self.status.link(case_obj.internal_id, sample["internal_id"])
                if link_obj:
                    link_obj.status = sample["status"] or link_obj.status
                    link_obj.mother = mother_obj or link_obj.mother
                    link_obj.father = father_obj or link_obj.father
                else:
                    new_link = self.status.relate_sample(
                        family=case_obj,
                        sample=family_samples[sample["name"]],
                        status=sample["status"],
                        mother=mother_obj,
                        father=father_obj,
                    )
                    self.status.add(new_link)
            self.status.add_commit(new_families)
        return new_families

    def store_samples(
        self,
        customer: str,
        order: str,
        ordered: dt.datetime,
        ticket: int,
        samples: List[dict],
    ) -> List[models.Sample]:
        """Store samples in the status database."""
        customer_obj = self.status.customer(customer)
        if customer_obj is None:
            raise OrderError(f"unknown customer: {customer}")
        new_samples = []

        with self.status.session.no_autoflush:
            for sample in samples:
                new_sample = self.status.add_sample(
                    name=sample["name"],
                    internal_id=sample["internal_id"],
                    sex=sample["sex"] or "unknown",
                    order=order,
                    ordered=ordered,
                    ticket=ticket,
                    priority=sample["priority"],
                    comment=sample["comment"],
                    tumour=sample["tumour"],
                    data_analysis=sample["data_analysis"],
                )
                new_sample.customer = customer_obj
                application_tag = sample["application"]
                application_version = self.status.current_application_version(application_tag)
                if application_version is None:
                    raise OrderError(f"Invalid application: {sample['application']}")
                new_sample.application_version = application_version
                new_samples.append(new_sample)

        self.status.add_commit(new_samples)
        return new_samples

    def store_fastq_samples(
        self,
        customer: str,
        order: str,
        ordered: dt.datetime,
        ticket: int,
        samples: List[dict],
    ) -> List[models.Sample]:
        """Store fast samples in the status database including family connection and delivery."""
        production_customer = self.status.customer("cust000")
        customer_obj = self.status.customer(customer)
        if customer_obj is None:
            raise OrderError(f"unknown customer: {customer}")
        new_samples = []

        with self.status.session.no_autoflush:
            for sample in samples:
                new_sample = self.status.add_sample(
                    name=sample["name"],
                    internal_id=sample["internal_id"],
                    sex=sample["sex"] or "unknown",
                    order=order,
                    ordered=ordered,
                    ticket=ticket,
                    priority=sample["priority"],
                    comment=sample["comment"],
                    tumour=sample["tumour"],
                    data_analysis=sample["data_analysis"],
                )
                new_sample.customer = customer_obj

                application_tag = sample["application"]
                application_version = self.status.current_application_version(application_tag)
                if application_version is None:
                    raise OrderError(f"Invalid application: {sample['application']}")
                new_sample.application_version = application_version
                new_samples.append(new_sample)

                if not new_sample.is_tumour:
                    new_family = self.status.add_family(
                        name=sample["name"], panels=["OMIM-AUTO"], priority="research"
                    )
                    new_family.customer = production_customer
                    self.status.add(new_family)

                    new_relationship = self.status.relate_sample(
                        family=new_family,
                        sample=new_sample,
                        status=sample["status"] or "unknown",
                    )
                    self.status.add(new_relationship)

                new_delivery = self.status.add_delivery(destination="caesar", sample=new_sample)
                self.status.add(new_delivery)

        self.status.add_commit(new_samples)
        return new_samples

    def store_microbial_order(
        self,
        customer: str,
        order: str,
        ordered: dt.datetime,
        ticket: int,
        lims_project: str,
        samples: List[dict],
        comment: str = None,
    ) -> models.MicrobialOrder:
        """Store microbial samples in the status database."""
        customer_obj = self.status.customer(customer)
        if customer_obj is None:
            raise OrderError(f"unknown customer: {customer}")
        with self.status.session.no_autoflush:
            new_order = self.status.add_microbial_order(
                customer=customer_obj,
                name=order,
                ordered=ordered,
                internal_id=lims_project,
                ticket_number=ticket,
                comment=comment,
            )
            for sample_data in samples:
                application_tag = sample_data["application"]
                application_version = self.status.current_application_version(application_tag)
                if application_version is None:
                    raise OrderError(f"Invalid application: {sample_data['application']}")

                organism = self.status.organism(sample_data["organism_id"])

                if not organism:
                    organism = self.status.add_organism(
                        internal_id=sample_data["organism_id"],
                        name=sample_data["organism_id"],
                        reference_genome=sample_data["reference_genome"],
                    )
                    self.status.add_commit(organism)

                new_sample = self.status.add_microbial_sample(
                    name=sample_data["name"],
                    internal_id=sample_data["internal_id"],
                    reference_genome=sample_data["reference_genome"],
                    comment=sample_data["comment"],
                    organism=organism,
                    application_version=application_version,
                    priority=sample_data["priority"],
                    data_analysis=sample_data["data_analysis"],
                )
                new_order.microbial_samples.append(new_sample)

        self.status.add_commit(new_order)
        return new_order

    def store_pools(
        self,
        customer: str,
        order: str,
        ordered: dt.datetime,
        ticket: int,
        pools: List[dict],
    ) -> List[models.Pool]:
        """Store pools in the status database."""
        customer_obj = self.status.customer(customer)
        if customer_obj is None:
            raise OrderError(f"unknown customer: {customer}")
        new_pools = []
        for pool in pools:
            with self.status.session.no_autoflush:
                application_version = self.status.current_application_version(pool["application"])
                if application_version is None:
                    raise OrderError(f"Invalid application: {pool['application']}")
            new_pool = self.status.add_pool(
                customer=customer_obj,
                name=pool["name"],
                order=order,
                ordered=ordered,
                ticket=ticket,
                application_version=application_version,
                data_analysis=pool["data_analysis"],
                capture_kit=pool["capture_kit"],
            )
            new_delivery = self.status.add_delivery(destination="caesar", pool=new_pool)
            self.status.add(new_delivery)
            new_pools.append(new_pool)
        self.status.add_commit(new_pools)
        return new_pools



"""Click commands to store microsalt analyses"""
import datetime as dt
import logging
from pathlib import Path

import click

from cg.apps import hk, tb
from cg.exc import AnalysisNotFinishedError, AnalysisDuplicationError
from cg.store import Store

LOG = logging.getLogger(__name__)


@click.group()
@click.pass_context
def store(context):
    """Store results from MIP in housekeeper."""
    context.obj["db"] = Store(context.obj["database"])
    context.obj["tb_api"] = tb.TrailblazerAPI(context.obj)
    context.obj["hk_api"] = hk.HousekeeperAPI(context.obj)


@store.command()
@click.argument("config-stream", type=click.File("r"), required=False)
@click.pass_context
def analysis(context, config_stream):
    """Store a finished analysis in Housekeeper."""
    status = context.obj["db"]
    tb_api = context.obj["tb_api"]
    hk_api = context.obj["hk_api"]

    if not config_stream:
        LOG.error("provide a config, suggestions:")
        for analysis_obj in context.obj["tb_api"].analyses(status="completed", deleted=False)[:25]:
            click.echo(analysis_obj.config_path)
        context.abort()

    new_analysis = _gather_files_and_bundle_in_housekeeper(
        config_stream, context, hk_api, status, tb_api
    )

    status.add_commit(new_analysis)
    click.echo(click.style("included files in Housekeeper", fg="green"))


def _gather_files_and_bundle_in_housekeeper(config_stream, context, hk_api, status, tb_api):
    """Function to gather files and bundle in housekeeper"""
    try:
        bundle_data = tb_api.add_analysis(config_stream)
    except AnalysisNotFinishedError as error:
        click.echo(click.style(error.message, fg="red"))
        context.abort()

    try:
        results = hk_api.add_bundle(bundle_data)
        if results is None:
            click.echo(click.style("analysis version already added", fg="yellow"))
            context.abort()
        bundle_obj, version_obj = results
    except FileNotFoundError as error:
        click.echo(click.style(f"missing file: {error.args[0]}", fg="red"))
        context.abort()

    family_obj = _add_new_analysis_to_the_status_api(bundle_obj, status)
    _reset_action_from_running_on_family(family_obj)
    new_analysis = _add_new_complete_analysis_record(bundle_data, family_obj, status, version_obj)
    version_date = version_obj.created_at.date()
    click.echo(f"new bundle added: {bundle_obj.name}, version {version_date}")
    _include_files_in_housekeeper(bundle_obj, context, hk_api, version_obj)

    return new_analysis


def _include_files_in_housekeeper(bundle_obj, context, hk_api, version_obj):
    """Function to include files in housekeeper"""
    try:
        hk_api.include(version_obj)
    except hk.VersionIncludedError as error:
        click.echo(click.style(error.message, fg="red"))
        context.abort()
    hk_api.add_commit(bundle_obj, version_obj)


def _add_new_complete_analysis_record(bundle_data, family_obj, status, version_obj):
    """Function to create and return a new analysis database record"""
    pipeline = family_obj.links[0].sample.data_analysis
    pipeline = pipeline if pipeline else "mip"  # TODO remove this default from here

    if status.analysis(family=family_obj, started_at=version_obj.created_at):
        raise AnalysisDuplicationError(
            f"Analysis object already exists for {family_obj.internal_id}{version_obj.created_at}"
        )

    new_analysis = status.add_analysis(
        pipeline=pipeline,
        version=bundle_data["pipeline_version"],
        started_at=version_obj.created_at,
        completed_at=dt.datetime.now(),
        primary=(len(family_obj.analyses) == 0),
    )
    new_analysis.family = family_obj
    return new_analysis


def _reset_action_from_running_on_family(family_obj):
    family_obj.action = None


def _add_new_analysis_to_the_status_api(bundle_obj, status):
    family_obj = status.family(bundle_obj.name)
    return family_obj


@store.command()
@click.pass_context
def completed(context):
    """Store all completed analyses."""
    hk_api = context.obj["hk_api"]
    for analysis_obj in context.obj["tb_api"].analyses(status="completed", deleted=False):
        existing_record = hk_api.version(analysis_obj.family, analysis_obj.started_at)
        if existing_record:
            LOG.debug("analysis stored: %s - %s", analysis_obj.family, analysis_obj.started_at)
            continue
        click.echo(click.style(f"storing family: {analysis_obj.family}", fg="blue"))
        with Path(analysis_obj.config_path).open() as config_stream:
            context.invoke(analysis, config_stream=config_stream)



"""The get_links command"""
import logging
from typing import List

import click

from cg.store import Store, models

LOG = logging.getLogger(__name__)


def get_links(
    store: Store, case_id: str = None, sample_id: str = None
) -> List[models.FamilySample]:
    """Get link objects for a SAMPLE_ID
    Args:
        case_id(str): petname
        sample_id(str): ACC6395A2
    Returns:
        link_objs(list): [models.FamilySample]
    """
    link_objs = []

    if case_id and sample_id:
        LOG.info("Link only one sample in a case")
        link_obj = store.link(family_id=case_id, sample_id=sample_id)
        if not link_obj:
            LOG.error("Could not find link for case %s and sample %s", case_id, sample_id)
            raise click.Abort
        link_objs = [link_obj]

    elif case_id:
        LOG.info("Link all samples in a case")
        case_obj = store.family(case_id)
        if not case_obj:
            LOG.error("Could not find case %s", case_id)
            raise click.Abort

        if not case_obj.links:
            LOG.error("Could not find links for case %s", case_id)
            raise click.Abort

        link_objs = case_obj.links

    elif sample_id:
        LOG.info("Link sample %s in all its families", sample_id)
        sample_obj = store.sample(sample_id)

        if not sample_obj:
            LOG.error(
                "Could not find sample %s. Did you intend %s as a case-id?",
                sample_id,
                sample_id,
            )
            raise click.Abort

        if not sample_obj.links:
            LOG.error("Could not find links for sample %s", sample_id)
            raise click.Abort

        link_objs = sample_obj.links

    else:
        LOG.error("Please provide case and/or sample")
        raise click.Abort

    return link_objs



""" Common microsalt related functionality """
from datetime import datetime
import logging
import re

from cg.store.models import Sample
from cg.apps.lims import LimsAPI

#    Organism - Fallback based on reference, ‘Other species’ and ‘Comment’. Default to “Unset”.
#    Priority = Default to empty string. Weird response. Typically “standard” or “research”.
#    Reference = Defaults to “None”
#    Method: Outputted as “1273:23”. Defaults to “Not in LIMS”
#    Date: Returns latest == most recent date. Outputted as DT object “YYYY MM DD”. Defaults to
#    datetime.min


class LimsMicrosaltAPI:
    """ Group microSALT specific functionality """

    def __init__(self, lims: LimsAPI, logger=logging.getLogger(__name__)):
        self.log = logger
        self.lims = lims

    def get_lims_comment(self, sample_id: str) -> str:
        """ returns the comment associated with a sample stored in lims"""
        lims_sample = self.lims.sample(sample_id)
        comment = lims_sample.get("comment") or ""
        if re.match(r"\w{4}\d{2,3}", comment):
            return comment

        return ""

    def get_organism(self, sample_obj: Sample) -> str:
        """Organism
        - Fallback based on reference, ‘Other species’ and ‘Comment’.
        Default to "Unset"."""

        organism = sample_obj.organism.internal_id.strip()
        comment = self.get_lims_comment(sample_obj.internal_id)
        has_comment = bool(comment)

        if "gonorrhoeae" in organism:
            organism = "Neisseria spp."
        elif "Cutibacterium acnes" in organism:
            organism = "Propionibacterium acnes"

        if organism == "VRE":
            reference = sample_obj.organism.reference_genome
            if reference == "NC_017960.1":
                organism = "Enterococcus faecium"
            elif reference == "NC_004668.1":
                organism = "Enterococcus faecalis"
            elif has_comment:
                organism = comment

        return organism

    def get_parameters(self, sample_obj: Sample):
        """Fill a dict with case config information for one sample """
        method_library_prep = self.lims.get_prep_method(sample_obj.internal_id)
        if method_library_prep:
            method_library_prep, _ = method_library_prep.split(" ", 1)
        method_sequencing = self.lims.get_sequencing_method(sample_obj.internal_id)
        if method_sequencing:
            method_sequencing, _ = method_sequencing.split(" ", 1)
        organism = self.get_organism(sample_obj)
        priority = "research" if sample_obj.priority == 0 else "standard"

        parameter_dict = {
            "CG_ID_project": sample_obj.microbial_order.internal_id,
            "Customer_ID_project": sample_obj.microbial_order.ticket_number,
            "CG_ID_sample": sample_obj.internal_id,
            "Customer_ID_sample": sample_obj.name,
            "organism": organism,
            "priority": priority,
            "reference": sample_obj.organism.reference_genome,
            "Customer_ID": sample_obj.microbial_order.customer.internal_id,
            "application_tag": sample_obj.application_version.application.tag,
            "date_arrival": str(sample_obj.received_at or datetime.min),
            "date_sequencing": str(sample_obj.sequenced_at or datetime.min),
            "date_libprep": str(sample_obj.prepared_at or datetime.min),
            "method_libprep": method_library_prep or "Not in LIMS",
            "method_sequencing": method_sequencing or "Not in LIMS",
        }

        return parameter_dict



"""Test methods for cg/cli/set/microbial_order"""

import pytest

from cg.cli.set import microbial_order
from cg.store import Store

SUCCESS = 0


def test_invalid_order_empty_db(cli_runner, base_context):
    """Test to create an order in a empty database"""
    # GIVEN an empty database

    # WHEN running set with an order that does not exist
    order_id = "dummy_order_id"
    application_tag = "dummy_application"
    result = cli_runner.invoke(
        microbial_order, [order_id, "sign", "-a", application_tag], obj=base_context
    )

    # THEN then it should complain on invalid order
    assert result.exit_code != SUCCESS


def test_invalid_order_non_empty_db(cli_runner, base_context, base_store: Store, helpers):
    """Test to create an non existing order in a populated database"""
    # GIVEN a non empty database
    order_id = "dummy_order_id"
    helpers.ensure_microbial_order(base_store, order_id=order_id)

    non_existing_order_id = "another_order_id"

    # WHEN running set with an order that does not exist
    application_tag = "dummy_application"
    result = cli_runner.invoke(
        microbial_order,
        [non_existing_order_id, "sign", "-a", application_tag],
        obj=base_context,
    )

    # THEN then it should complain on invalid order
    assert result.exit_code != SUCCESS


def test_valid_order_no_apptag_option(cli_runner, base_context, base_store: Store, helpers):
    """Test to create an order without an apptag"""
    # GIVEN a non empty database
    order = helpers.ensure_microbial_order(base_store)

    # WHEN running set with an order that does not exist
    order_id = order.internal_id
    result = cli_runner.invoke(microbial_order, [order_id, "sign"], obj=base_context)

    # THEN then it should just exit
    assert result.exit_code != SUCCESS


@pytest.mark.parametrize("option_key", ["--name", "--ticket"])
def test_set_option(cli_runner, base_context, base_store: Store, option_key, helpers):
    """Test to set an order with an option"""

    # GIVEN a database with an order
    order = helpers.ensure_microbial_order(base_store)
    option_value = "option_value"

    # WHEN calling set_microbial_order with option
    signature = "sign"
    result = cli_runner.invoke(
        microbial_order,
        [order.internal_id, signature, option_key, option_value],
        obj=base_context,
    )

    # THEN then the option should have been set on the object and the user been informed
    assert result.exit_code == SUCCESS
    assert order.internal_id in result.output
    assert option_value in result.output
    assert option_value in base_store.MicrobialOrder.query.first().to_dict().values()
    assert signature in base_store.MicrobialOrder.query.first().comment


def test_set_project_name_in_lims(cli_runner, base_context, base_store, helpers):
    """Test to set an order with a option so that LIMS info is outputed"""

    # GIVEN a database with an order
    order = helpers.ensure_microbial_order(base_store)
    name = "a_name"

    # WHEN calling set_microbial_order with option
    signature = "sign"
    result = cli_runner.invoke(
        microbial_order,
        [order.internal_id, signature, "--name", name],
        obj=base_context,
    )

    # THEN then the option should have been set on the object and the user been informed
    assert result.exit_code == SUCCESS
    assert "LIMS" in result.output
    assert name in result.output
    assert name in base_context["lims"].get_updated_project_name()


@pytest.mark.parametrize("option_key", ["--name", "--ticket"])
def test_set_option_with_same_value(
    cli_runner, base_context, base_store: Store, option_key, helpers
):
    """Test to set an order with the same option_value twice"""

    # GIVEN a database with an order
    order = helpers.ensure_microbial_order(base_store)
    option_value = "option_value"
    first_sign = "first_sign"
    second_sign = "second_sign"

    # WHEN calling set_microbial_order with same option_value
    cli_runner.invoke(
        microbial_order,
        [order.internal_id, first_sign, option_key, option_value],
        obj=base_context,
    )
    result = cli_runner.invoke(
        microbial_order,
        [order.internal_id, second_sign, option_key, option_value],
        obj=base_context,
    )

    # THEN then the second call should not render a change in comment
    assert result.exit_code == SUCCESS
    assert order.internal_id in result.output
    assert option_value in result.output
    assert option_value in base_store.MicrobialOrder.query.first().to_dict().values()
    assert first_sign in base_store.MicrobialOrder.query.first().comment
    assert second_sign not in base_store.MicrobialOrder.query.first().comment


def test_old_comment_preserved(cli_runner, base_context, base_store: Store, helpers):
    """Test to set an order and check if old comment is preserved"""

    # GIVEN a database with an order
    order = helpers.ensure_microbial_order(base_store)
    first_sign = "first_sign"
    second_sign = "second_sign"

    # WHEN calling set_microbial_order twice with different signatures
    cli_runner.invoke(
        microbial_order,
        [order.internal_id, first_sign, "--name", "dummy_name1"],
        obj=base_context,
    )
    result = cli_runner.invoke(
        microbial_order,
        [order.internal_id, second_sign, "--name", "dummy_name2"],
        obj=base_context,
    )

    # THEN both signatures should be found in the comment
    assert result.exit_code == SUCCESS
    assert first_sign in base_store.MicrobialOrder.query.first().comment
    assert second_sign in base_store.MicrobialOrder.query.first().comment



"""Add CLI support to start microsalt"""

import json
import logging
from pathlib import Path
import subprocess

import click

from cg.apps import hk, lims
from cg.apps.usalt.fastq import FastqHandler
from cg.cli.workflow.microsalt.store import store as store_cmd
from cg.cli.workflow.microsalt.deliver import (
    deliver as deliver_cmd,
    PROJECT_TAGS,
    SAMPLE_TAGS,
)
from cg.meta.microsalt.lims import LimsMicrosaltAPI
from cg.meta.workflow.microsalt import AnalysisAPI
from cg.meta.deliver import DeliverAPI
from cg.store import Store

LOG = logging.getLogger(__name__)


@click.group("microsalt", invoke_without_command=True)
@click.option("-o", "--order", "order_id", help="include all microbial samples for an order")
@click.pass_context
def microsalt(context: click.Context, order_id):
    """Microbial workflow"""
    context.obj["db"] = Store(context.obj["database"])
    hk_api = hk.HousekeeperAPI(context.obj)
    lims_api = lims.LimsAPI(context.obj)
    deliver = DeliverAPI(
        context.obj,
        hk_api=hk_api,
        lims_api=lims_api,
        case_tags=PROJECT_TAGS,
        sample_tags=SAMPLE_TAGS,
    )
    context.obj["api"] = AnalysisAPI(db=context.obj["db"], hk_api=hk_api, lims_api=lims_api)
    context.obj["lims_microsalt_api"] = LimsMicrosaltAPI(lims=lims_api)

    if context.invoked_subcommand is None:
        if order_id is None:
            LOG.error("Please provide an order")
            context.abort()
        else:
            # execute the analysis!
            context.invoke(config_case, order_id=order_id)
            context.invoke(link, order_id=order_id)
            context.invoke(run, order_id=order_id)


@microsalt.command()
@click.option("-o", "--order", "order_id", help="link all microbial samples for an order")
@click.argument("sample_id", required=False)
@click.pass_context
def link(context: click.Context, order_id: str, sample_id: str):
    """Link microbial FASTQ files for a SAMPLE_ID"""
    sample_objs = None

    if order_id and (sample_id is None):
        # link all samples in a case
        sample_objs = context.obj["db"].microbial_order(order_id).microbial_samples
    elif sample_id and (order_id is None):
        # link sample in all its families
        sample_objs = [context.obj["db"].microbial_sample(sample_id)]
    elif sample_id and order_id:
        # link only one sample in a case
        sample_objs = [context.obj["db"].microbial_sample(sample_id)]

    if not sample_objs:
        LOG.error("provide order and/or sample")
        context.abort()

    for sample_obj in sample_objs:
        LOG.info("%s: link FASTQ files", sample_obj.internal_id)
        context.obj["api"].link_sample(
            FastqHandler(context.obj),
            case=sample_obj.microbial_order.internal_id,
            sample=sample_obj.internal_id,
        )


@microsalt.command("config-case")
@click.option("-d", "--dry", is_flag=True, help="print config-case to console")
@click.option(
    "-o",
    "--order",
    "order_id",
    help="create config-case all microbial samples for an order",
)
@click.argument("sample_id", required=False)
@click.pass_context
def config_case(context: click.Context, dry, order_id: str, sample_id: str):
    """ Create a config file on case level for microSALT """
    if order_id and (sample_id is None):
        microbial_order_obj = context.obj["db"].microbial_order(order_id)
        if not microbial_order_obj:
            LOG.error("Order %s not found", order_id)
            context.abort()
        sample_objs = microbial_order_obj.microbial_samples
    elif sample_id and (order_id is None):
        sample_obj = context.obj["db"].microbial_sample(sample_id)
        if not sample_obj:
            LOG.error("Sample %s not found", sample_id)
            context.abort()
        sample_objs = [context.obj["db"].microbial_sample(sample_id)]
    elif sample_id and order_id:
        microbial_order_obj = context.obj["db"].microbial_order(order_id)
        if not microbial_order_obj:
            LOG.error("Samples %s not found in %s ", sample_id, order_id)
            context.abort()
        sample_objs = [
            sample_obj
            for sample_obj in microbial_order_obj.microbial_samples
            if sample_obj.internal_id == sample_id
        ]
    else:
        LOG.error("provide order and/or sample")
        context.abort()

    parameters = [
        context.obj["lims_microsalt_api"].get_parameters(sample_obj) for sample_obj in sample_objs
    ]

    filename = order_id if order_id else sample_id
    outfilename = Path(context.obj["usalt"]["queries_path"]) / filename
    outfilename = outfilename.with_suffix(".json")
    if dry:
        print(json.dumps(parameters, indent=4, sort_keys=True))
    else:
        with open(outfilename, "w") as outfile:
            json.dump(parameters, outfile, indent=4, sort_keys=True)


@microsalt.command()
@click.option("-d", "--dry", is_flag=True, help="print command to console")
@click.option("-c", "--config-case", required=False, help="optionally change the config-case")
@click.argument("order_id")
@click.pass_context
def run(context, dry, config_case, order_id):
    """ Start microSALT with an order_id """
    microsalt_command = context.obj["usalt"]["binary_path"]
    command = [microsalt_command]

    config_case_path = config_case
    if not config_case:
        queries_path = Path(context.obj["usalt"]["queries_path"])
        config_case_path = queries_path / order_id
        config_case_path = config_case_path.with_suffix(".json")

    command.extend(["--parameters", str(config_case_path)])
    if dry:
        print(" ".join(command))
    else:
        LOG.info("Starting microSALT! '%s'", " ".join(command))
        subprocess.run(command, shell=True, check=True)


microsalt.add_command(config_case)
microsalt.add_command(deliver_cmd)
microsalt.add_command(run)
microsalt.add_command(store_cmd)
